#!perl -w
# ({[
# $Id: nwsc.pl,v 1.10 2007-02-06 20:41:29 tsbirn Exp $
#-------------------------------------------------------------------------------
# nwsc.pl - Win32 service running a tiny httpd to execute WMI-Queries
# to Windows-PCs. The queries are initialized by a nagios server's plugin.
# See also:
#    Windows::Daemon    at http://www.roth.net
#    p5httpd and phttpd at http://www.pault.com
#    Forking            at http://use.perl.org/~jjohn/journal/18668
#    Multithreading     at http://burks.bton.ac.uk/burks/language/perl/perlman/pod/perlipc/tcpcs.htm
# Authors: Hermann Gottschalk (hg@ostc.de) at http://www.ostc.de
#          Thomas Birnthaler  (tb@ostc.de) at http://www.ostc.de
# Project: Monitoring Bionorica AG Neumarkt http://www.bionorica.de
# (C) 2006 OSTC Open Source Training and Consulting GmbH
#-------------------------------------------------------------------------------
use strict;

use vars qw( $VERSION $NAME $DATE );
use Getopt::Long;
use Socket;
use Win32;
use Win32::Daemon;
use Win32::OLE qw(in with);
use Win32::Process;
use POSIX ":sys_wait_h";
###use POSIX ":sys_wait_h SIGALARM";
my %children = ();
my $MAX_KIDS = 10;

# Standardausgabe im Binrmodus ffnen (d.h. \n nicht nach \r\n umsetzen)
binmode(STDOUT);

# Signale unsicher machen (Perl fngt sie nicht mehr ab)
$ENV{PERL_SIGNALS} = "unsafe";
my $wmi_timeout = 25;

# Constants
my $STATE_OK       = 0;
my $STATE_WARNING  = 1;
my $STATE_CRITICAL = 2;
my $STATE_UNKNOWN  = -1;

my $DATE;
my $VERSION = '$Id: nwsc.pl,v 1.10 2007-02-06 20:41:29 tsbirn Exp $';
my $NAME    = '$Name:  $';

$NAME    =~ s/^\$Name:  $$/$1/g;
$VERSION =~ m#^.*,v\ ([\d\.]+)\ ([\d/]+)\ .*$#;
($VERSION, $DATE) = ($1, $2);

my $listener = 0;
my $listener_pid = -1;

my $paddr;
my $PrevState;
my $State;
my $thispid;
my $localname;
my $mime_type = 'text/html';
my %Config = (
		port    => 80,
		logfile => "C:\\WinNT\\Temp\\nwsc.log",
);
my %communicators;

# forward declarations
sub logerr($$);
sub logmsg($);

Getopt::Long::Configure( "prefix_pattern=(-|\/)" );
my $Result = GetOptions( \%Config,
			qw(
				install|i
				remove|r
				user|u=s
				password|g=s
				logfile|l=s
				port|p=i
				maxkids|m=i
				help|?|h
				listen|t
				communicate|c
			));

my $port = $Config{port};
$MAX_KIDS = $Config{maxkids} if defined($Config{maxkids});


$Config{help} = 1 if ( ! $Result or scalar @ARGV );

if ( $Config{install} )
{
	&install();
	exit();
}
elsif ( $Config{remove} )
{
	&remove();
	exit();
}
elsif ( $Config{listen} )
{
	&listen_for_request();
	exit();
}
elsif ( $Config{communicate} )
{
	&handle_request();
	exit();
}
elsif ( $Config{help} )
{
	&syntax();
	exit();
}
else
{
	&logmsg("NOTHING TO DO");
}


# Start the service...
if ( Win32::Daemon::StartService() )
{
	&logmsg("# Software: $0\n");
	&logmsg("# Date: " . localtime() . "\n");
}
else
{
	&logmsg("Failed to start this script as a Win32 service.\n");
	&logmsg("Error: " . &get_error("DAEMON") . "\n");
	exit();
}

while ( SERVICE_STOPPED != ( $State = Win32::Daemon::State() ) )
{
	if ( SERVICE_START_PENDING == $State )
	{
		# Initialization code
		Win32::Daemon::State( SERVICE_RUNNING );
		$PrevState = SERVICE_RUNNING;
	}
	elsif ( SERVICE_PAUSE_PENDING == $State )
	{
		# "Pausing...";
		Win32::Daemon::State( SERVICE_PAUSED );
		$PrevState = SERVICE_PAUSED;
		next;
	}
	elsif ( SERVICE_CONTINUE_PENDING == $State )
	{
		# "Resuming...";
		Win32::Daemon::State( SERVICE_RUNNING );
		$PrevState = SERVICE_RUNNING;
		next;
	}
	elsif ( SERVICE_STOP_PENDING == $State )
	{
		# "Stopping...";
		Win32::Daemon::State( SERVICE_STOPPED );
		$PrevState = SERVICE_STOPPED;
		next;
	}
	elsif ( SERVICE_RUNNING == $State )
	{
		unless ($listener)
		{
			&logmsg("Starting listener");
			$listener_pid = &start_listener();
			&logmsg("Listener PID $listener_pid");
			$listener = 1;
		}
		sleep(1);
		$PrevState = SERVICE_RUNNING;
	}
	else
	{
		# We have some unknown state...
		# reset it back to what we last knew the state to be...
		Win32::Daemon::State( $PrevState );
		sleep( $Config{timeout_value} );
	}
}

&cleanup_kids;
&logmsg("Stopping Service");
Win32::Daemon::StopService();

&logmsg("Service stopped");

################################## Subroutines ###################

sub get_service_config
{
	my $ScriptPath = join( "", Win32::GetFullPathName( $0 ) );
	my %Hash = (
		name       => 'NWSC',
		display    => 'Nagios WindowsService-Checker Service',
		path       => $^X,
		user       => "$Config{user}",       # TODO: fill if -u not given
		password   => "$Config{password}",   # TODO: fill if -p not given
		description   => "Checks Windows Services on RemoteHosts - OSTC GmbH",
		parameters => "$ScriptPath -l \"$Config{logfile}\" -p $Config{port} -m $MAX_KIDS",
	);
	return( \%Hash );
}

sub install
{
	my $ServiceConfig = &get_service_config();

	if ( Win32::Daemon::CreateService( $ServiceConfig ) )
	{
		print "The $ServiceConfig->{display} was successfully installed.\n";
	}
	else
	{
		print "Failed to add the $ServiceConfig->{display} service.\nError: " . &get_error("DAEMON") . "\n";
	}
}

sub remove
{
	my $ServiceConfig = &get_service_config();

	if ( Win32::Daemon::DeleteService( $ServiceConfig->{name} ) )
	{
		print "The $ServiceConfig->{display} was successfully removed.\n";
	}
	else
	{
		print "Failed to remove the $ServiceConfig->{display} service.\nError: " . &get_error("DAEMON") . "\n";
	}
}

sub dump_error
{
	print &get_error("DAEMON"), "\n";
}

sub get_error
{
	my $type = shift;

	if ($type eq "OLE") {
		return( Win32::FormatMessage( Win32::Daemon::GetLastError() ) );
	}
	elsif ($type eq "DAEMON") {
		return( Win32::FormatMessage( Win32::Daemon::GetLastError() ) );
	}
}

sub start_listener
{
	my $listener;
	my $pgmpath = $0;
	my ($basedir, $pgm) = Win32::GetFullPathName( $0 );
	
	Win32::Process::Create(
		$listener,
		$^X,   ###  "C:\\Perl\\bin\\perl.exe",
		"$^X $0 -listen -port $Config{port} -logfile $Config{logfile} -m $MAX_KIDS",
		0,
		HIGH_PRIORITY_CLASS,     ### | CREATE_NEW_CONSOLE, ### DETACHED_PROCESS,
		$basedir,
	) or &logmsg(Win32::FormatMessage(Win32::GetLastError())); 
	&logmsg("Listener Started"); 

	return $listener->GetProcessID();
}

sub start_communicator
{
	my $communicator;
	my $pgmpath = $0;
	my ($basedir, $pgm) = Win32::GetFullPathName( $0 );
	
	Win32::Process::Create(
		$communicator,
		$^X,   ###  "C:\\Perl\\bin\\perl.exe",
		"$^X $0 -communicate -port $Config{port} -logfile $Config{logfile}",
		1,
		HIGH_PRIORITY_CLASS,     ### | CREATE_NEW_CONSOLE, ### DETACHED_PROCESS,
		$basedir,
	) or &logmsg("XXX" . Win32::FormatMessage(Win32::GetLastError())); 
	&logmsg("Communicator Started"); 

	return $communicator->GetProcessID();
}

sub listen_for_request
{
	my $tcp = getprotobyname('tcp');
	my $communicator_pid;

	socket(Server, PF_INET, SOCK_STREAM, $tcp)      or &logmsg("socket: $!");
	setsockopt(Server, SOL_SOCKET, SO_REUSEADDR, pack("l", 1)) or warn "setsockopt: $!";
	bind(Server, sockaddr_in($port, INADDR_ANY))    or &logmsg("bind: $!");
	listen(Server,SOMAXCONN)                        or &logmsg("listen: $!");
	&logmsg("server started on port $port");

	while ($paddr = accept(Client,Server))
	{
		my $pobj;

		open(STDIN,  "<&Client") or &logmsg("Cannot copy Client1");
		open(STDOUT, ">&Client") or &logmsg("Cannot copy Client2");

		$communicator_pid = &start_communicator;
		$communicators{$communicator_pid} = 1;
		&logmsg("Communicator PID $communicator_pid");

		while (1) {
			foreach (keys %communicators) {
				delete $communicators{$_}, &logmsg("Gone $_") if (Win32::Process::Open($pobj, $_, 0) == 0);
			}
			last if (keys %communicators < $MAX_KIDS);
			&logmsg("Too much processes (" . (scalar keys %communicators) . ") waiting 1s...");
			sleep 1;
		}

		close STDIN;
		close STDOUT;
		close Client;
	}
	&logmsg("Fatal error: accept failed: $!\n"); # This should never happen
}

#################### other subroutines ####################

sub handle_request
{
	my $remote_sockaddr  = getpeername(STDIN);
	my (undef, $iaddr)   = sockaddr_in($remote_sockaddr);
	my $peername         = (gethostbyaddr($iaddr, AF_INET) or "localhost");
	my $peeraddr         = (inet_ntoa($iaddr) or "127.0.0.1");

	my $local_sockaddr   = getsockname(STDIN);
	   (undef, $iaddr)   = sockaddr_in($local_sockaddr);
	$localname           = (gethostbyaddr($iaddr, AF_INET) or "localhost");
	my $localaddr        = (inet_ntoa($iaddr) or "127.0.0.1");

	chomp($_ = <STDIN>);
	my ($method, $url, $proto, undef) = split;

	$url =~ s#\\#/#g;
	&logmsg("<- $peername: $_");
	my ($file, undef, $query_string) = ($url =~ /([^?]*)(\?(.*))?/); # split at ?
	my $file_escaped = $file;
	$file =~ s/%([\dA-Fa-f]{2})/chr(hex($1))/eg; # %20 -> space
	$query_string =~ s/%([\dA-Fa-f]{2})/chr(hex($1))/eg; # %20 -> space

	if ( $method !~ /^GET$/ ) {
		&logerr(400, "I don't understand method $method");
	}
	elsif ( "$file" eq "/check_me" ) {
		&logerr(200, "I am fine");
	}
	elsif ( "$file" ne "/check_service" ) {
		&logerr(400, "$file: $!");
	}
	else
	{
		my @env_vars = qw(USER_AGENT CONTENT_LENGTH CONTENT_TYPE);
		foreach my $var (@env_vars) {
			$ENV{$var} = ""; # delete $ENV{$var} will crash perl on Netbook :-()
		}
		while (<STDIN>)
		{
			s/[\r\l\n\s]+$//;
			/^User-Agent: (.+)/i      and $ENV{USER_AGENT} = $1;
			/^Content-length: (\d+)/i and $ENV{CONTENT_LENGTH} = $1;
			/^Content-type: (.+)/i    and $ENV{CONTENT_TYPE} = $1;
			last if (/^$/);
		}

		$ENV{SERVER_PROTOCOL} = $proto;
		$ENV{SERVER_PORT}     = $port;
		$ENV{SERVER_NAME}     = $localname;
		$ENV{SERVER_URL}      = "http://$localname:$port/";
		$ENV{SCRIPT_NAME}     = $file;
		$ENV{SCRIPT_FILENAME} = $file;
		$ENV{REQUEST_URI}     = $url;
		$ENV{REQUEST_METHOD}  = $method;
		$ENV{REMOTE_ADDR}     = $peeraddr;
		$ENV{REMOTE_HOST}     = $peername;
		$ENV{QUERY_STRING}    = $query_string;
		$ENV{SERVER_SOFTWARE} = "NWSC/$NAME";

		&check_service($query_string);
	}
	close Client;
}

sub logmsg ($)
{
	unless ( $Config{logfile} eq "" )
	{
		my $fulltime = localtime();
		my ($hms) = ($fulltime =~ /(\d\d:\d\d:\d\d)/);

		open( LOG, ">> $Config{logfile}" );
		print LOG  "$$ $hms @_\n";
		close LOG;
	}
}

sub logerr ($$)
{
	my ($code, $detail) = @_;
	my %codes = (
		200  => 'OK',
		400  => 'Bad Request',
		404  => 'Not Found',
		500  => 'Internal Server Error',
		501  => 'Not Implemented',
	);
	my $msg = "$code " . $codes{$code};
	my ($content, $length);

	&logmsg("-> $msg $detail");

	$content =  "<html><title>NWSC - Nagios Windows Service Checker ($NAME)</title><body>";
	$content .= "<h1>$msg</h1>";
	$content .= "<p>$detail</p>";
	$content .= "<hr>";
	$content .= "<p><I>NWSC/$NAME server at $localname port $port</I></p>";
	$content .= "<p>Use $localname/check_service?method=&lt;disk|disks|service|uptime|wmiclass&gt;&amp;host=&lt;hostname&gt;&amp;service=&lt;service(s)&gt;</p>";
	$content .= "</body></html>";
	$length = length($content);

	print "HTTP/1.0 $msg\n";
	print "Content-type: $mime_type\n";
	print "Content-length: $length\n";
	print "\n";
	print "$content";
}

sub check_service
{
	$_ = shift;
	my $state            = $STATE_UNKNOWN;
	my $state_of_service = $STATE_UNKNOWN;
	my $detail           = "";
	my @services;
	my ($method, $remotehost, $service, $services, $wmiclass);

	($method, $remotehost) = /^method=(disk|disks|service|uptime|wmiclass)&host=(.+)&.*$/ or &logerr(400, "Bad Query: $_");
	if ($method eq "service") {
		($services) = /^method=.+&host=.+&service=(.+)$/ or &logerr(400, "Bad Query: @_");
		(@services) = split /,/ , $services;
	}
	elsif ($method eq "wmiclass") {
		(undef, $wmiclass) = /^method=.+&host=.+&(service|wmiclass)=(.+)$/ or &logerr(400, "Bad Query: @_");
	}

	# WMI object, services and attributes
	my $WMI;
	my $rem_info;
	my $the_service;
	my $service_found = 0;
	my $msg_part;

	# Create a new WMI and get all Services from remote host
	$WMI = Win32::OLE->new('WbemScripting.SWbemLocator') or
		&resultmsg($STATE_CRITICAL, "Cannot access WMI on local machine: " . &get_error("OLE"));

	# POSIX-Signale verwenden
###	my $mask      = POSIX::SigSet->new( SIGALRM );
###	my $action    = POSIX::SigAction->new(sub { $logmsg("alarm raised"); exit(); }, $mask);
###	my $oldaction = POSIX::SigAction->new();
###	POSIX::sigaction(SIGALRM, $action, $oldaction );

	eval {
		local $SIG{ALRM} = sub { &logmsg("alarm raised"); exit(); };

		&logmsg("setting timout $wmi_timeout s");
		alarm($wmi_timeout);

		# Get all services on remote machine
		$rem_info = $WMI->ConnectServer($remotehost) or
			&resultmsg($STATE_CRITICAL, "Cannot access WMI on remote machine ($remotehost): " . &get_error("OLE"));
#		$rem_info = $WMI->ConnectServer($remotehost,
#			"root\\cimv2", "", "", "", "", 128
#		) or
#			&resultmsg($STATE_CRITICAL, "Cannot access WMI on remote machine ($remotehost): " . &get_error("OLE"));
#					(
#						"strServer"      => $remotehost,
#						"iSecurityFlags" => 0x80,
#					)
		alarm(0);

		if ( $method eq "service" ) {
			foreach (@services) {
				($state_of_service, $msg_part) = &check_for_service($_, \$rem_info);
				$detail .= "<b>" . $msg_part . "</b><br>";
				if ($state_of_service > $state) {$state = $state_of_service}
			}
		} elsif ($method eq "disk") {
			&check_for_disk();
		} elsif ($method eq "disks") {
			&check_for_disks();
		} elsif ($method eq "uptime") {
			&check_for_uptime();
		} elsif ($method eq "wmiclass") {
			$state = 0;
			$detail = "\n" . join("<br>\n", &check_for_wmiclass($wmiclass, \$rem_info));
			$detail .= "\n";
		}
		&resultmsg($state, $detail);
	}
}

sub check_for_disk
{
	# TODO
	return($STATE_OK, "DUMMY");
}

sub check_for_disks
{
	# TODO
	return($STATE_OK, "DUMMY");
}

sub check_for_uptime
{
	# TODO
	return($STATE_OK, "DUMMY");
}

sub check_for_wmiclass
{
	my ($wmiclass, $rem_info) = @_;
	my $rem_wmiclass;
	my $rem_wmiclasses = $$rem_info->InstancesOf($wmiclass);
	my @collect = ();

	for $rem_wmiclass (in($rem_wmiclasses))
	{
		foreach my $object (in($rem_wmiclass->{'Properties_'})) {
			if (ref($object->{Value}) eq "ARRAY") {
				push @collect, "  $object->{Name} = { ";
				foreach my $value (in($object->{Value}) ) {
					push @collect, "$value ";
				    
				}
				push @collect, "}";
			} else {
				push @collect, "  $object->{Name} = $object->{Value}";
			}
		}
		push @collect, "ENDOFCLASS";
	}
	return @collect;
}

sub check_for_service
{
	my ($the_service, $rem_info) = @_;
	my $rem_service;
	my $service_found = 0;
	my $rem_services = $$rem_info->InstancesOf("Win32_Service");

	for $rem_service (in($rem_services))
	{
		if ($rem_service->{"Name"} eq $the_service)
		{
			$the_service = $rem_service;
			$service_found = 1;
		}
	}

	if ($service_found)
	{
		return(&check_the_service(\$the_service));
	}
	else
	{
		return($STATE_WARNING, "Service $the_service not found");
	}
}

sub check_the_service
{
	my ($the_service) = (@_);
	my $state  = $$the_service->{"State"};
	my $status = $$the_service->{"Status"};
	my $name   = $$the_service->{"Name"};
	if ($state eq "Running") {
		return($STATE_OK, "$state-$name");
	}
	if ($state eq "Stopped") {
		return($STATE_CRITICAL, "$state-$name");
	}
	if ($state eq "Paused") {
		return($STATE_WARNING, "$state-$name");
	}
	if ($state eq "Start Pending") {
		return($STATE_WARNING, "$state-$name");
	}
	if ($state eq "Stop Pending") {
		return($STATE_WARNING, "$state-$name");
	}
	if ($state eq "Continue Pending") {
		return($STATE_WARNING, "$state-$name");
	}
	if ($state eq "Unknown") {
		return($STATE_CRITICAL, "$state-$name");
	}
	return($STATE_CRITICAL, "$name\' state: $state; status: $status");
}

sub resultmsg
{
	my ($state, $msg) = @_;
	my ($content, $length);

	&logmsg("$state: $msg");

	$content = "<html>";
	$content .= "<title>NWSC - Nagios Windows Service Checker ($NAME)</title>";
	$content .= "<body>Result: Service(s) ";

	if ( $state == 0 ) {
		$content .= "Okay<br> $msg";
	}
	elsif ( $state == 1 ) {
		$content .= "Warning<br> $msg";
	}
	else {
		$content .= "Critical<br> $msg";
	}

	$content .= "</body></html>";
	$length = length($content);

	print "HTTP/1.0 200 OK\n";
	print "Content-type: $mime_type\n";
	print "Content-length: $length\n";
	print "\n";
	print "$content";
}

sub cleanup_kids
{
	logmsg("Cleaning listener");
	Win32::Process::KillProcess($listener_pid, 0);
}

sub syntax
{
	my( $Script ) = ( $0 =~ m#([^\\/]+)$# );
	my $Line = "-" x length( $Script );
	print << "EOT";

  $Script
  $Line
  Win32 service that starts a tiny httpd for WMI-requests by nagios
  Version: $NAME
  Syntax:
    $0 [OPTS]

    -l/-logfile <PATH>......Path to store log file, empty string ("")
                            disables logging (default: $Config{logfile})
    -p/-port <PORT>.........Httpd port (default: 80)
    -m/maxkids <NUMBER>.....Number of subprocesses allowed (default: $MAX_KIDS)
    -u/user <USERNAME>......Administrative user of remote hosts
    -g/password <PASSWORD>..Password of administrative user of remote hosts
    -install................Install script as a Win32 service,
                            additional parameters passed in will set 
                            when the script runs as a service.
    -remove.................Remove script as a Win32 service.

  Examples:
    perl $0 -install
    perl $0 -install -l c:\\checker.log -p 6500
    perl $0 -install -l c:\\checker.log -p 6500 -u NORGE\\gatesb -g t0p\$ecret
    perl $0 -install -l c:\\checker.log -p 6500 -m 50
    perl $0 -remove

  (C) 2006 OSTC GmbH, Hermann Gottschalk, Thomas Birnthaler (nwsc(at)ostc.de)
  http://www.ostc.de/

EOT
}

__END__
# ]})
